from __future__ import annotations

import base64
import hashlib
import mimetypes
import tempfile
import sys
import zipfile
from enum import Enum
from http.cookiejar import CookieJar
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, BinaryIO, Optional, Sequence
from requests import Response
from requests.auth import AuthBase
from requests.models import PreparedRequest
import msgspec
import requests
from architecture.utils.decorators import ensure_module_installed
from typing import (
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    TypeAlias,
    overload,
    Literal,
)
from typing_extensions import Self


if TYPE_CHECKING:
    from fastapi import UploadFile as FastAPIUploadFile
    from _typeshed import SupportsItems, SupportsRead
    from litestar.datastructures import UploadFile as LitestarUploadFile

    _TextMapping: TypeAlias = MutableMapping[str, str]
    _HeadersMapping: TypeAlias = Mapping[str, str | bytes | None]

    _Data: TypeAlias = (
        # used in requests.models.PreparedRequest.prepare_body
        #
        # case: is_stream
        # see requests.adapters.HTTPAdapter.send
        # will be sent directly to http.HTTPConnection.send(...) (through urllib3)
        Iterable[bytes]
        # case: not is_stream
        # will be modified before being sent to urllib3.HTTPConnectionPool.urlopen(body=...)
        # see requests.models.RequestEncodingMixin._encode_params
        # see requests.models.RequestEncodingMixin._encode_files
        # note that keys&values are converted from Any to str by urllib.parse.urlencode
        | str
        | bytes
        | SupportsRead[str | bytes]
        | list[tuple[Any, Any]]
        | tuple[tuple[Any, Any], ...]
        | Mapping[Any, Any]
    )

    _ParamsMappingKeyType: TypeAlias = str | bytes | int | float
    _ParamsMappingValueType: TypeAlias = (
        str | bytes | int | float | Iterable[str | bytes | int | float] | None
    )
    _Params: TypeAlias = (
        SupportsItems[_ParamsMappingKeyType, _ParamsMappingValueType]
        | tuple[_ParamsMappingKeyType, _ParamsMappingValueType]
        | Iterable[tuple[_ParamsMappingKeyType, _ParamsMappingValueType]]
        | str
        | bytes
    )
    _Verify: TypeAlias = bool | str
    _Timeout: TypeAlias = float | tuple[float, float] | tuple[float, None]
    _Cert: TypeAlias = str | tuple[str, str]
    Incomplete: TypeAlias = Any
    _Hook: TypeAlias = Callable[[Response], Any]
    _HooksInput: TypeAlias = Mapping[str, Iterable[_Hook] | _Hook]
    _FileContent: TypeAlias = SupportsRead[str | bytes] | str | bytes
    _FileName: TypeAlias = str | None
    _FileContentType: TypeAlias = str
    _FileSpecTuple2: TypeAlias = tuple[_FileName, _FileContent]
    _FileSpecTuple3: TypeAlias = tuple[_FileName, _FileContent, _FileContentType]
    _FileCustomHeaders: TypeAlias = Mapping[str, str]
    _FileSpecTuple4: TypeAlias = tuple[
        _FileName, _FileContent, _FileContentType, _FileCustomHeaders
    ]
    _FileSpec: TypeAlias = (
        _FileContent | _FileSpecTuple2 | _FileSpecTuple3 | _FileSpecTuple4
    )
    _Files: TypeAlias = Mapping[str, _FileSpec] | Iterable[tuple[str, _FileSpec]]
    _Auth: TypeAlias = (
        tuple[str, str] | AuthBase | Callable[[PreparedRequest], PreparedRequest]
    )


class FileExtension(str, Enum):
    PDF = "pdf"
    JSON = "json"
    PNG = "png"
    JPEG = "jpeg"
    JPG = "jpg"
    HTML = "html"
    TXT = "txt"
    MD = "md"
    ZIP = "zip"
    DOCX = "docx"
    XLSX = "xlsx"
    CSV = "csv"
    PPTX = "pptx"
    TIFF = "tiff"
    # Add more extensions as needed


class RawFile(msgspec.Struct, frozen=True, gc=False):
    """
    Represents an immutable raw file with its content and extension.

    The `RawFile` class is designed for efficient and immutable handling of raw file data.
    It stores file contents as immutable bytes and provides utility methods for reading,
    writing, and manipulating the file content without mutating the original data.

    **Key Features:**

    - **Immutability**: Instances of `RawFile` are immutable. Once created, their contents cannot be modified.
      This is enforced by using `msgspec.Struct` with `frozen=True`, ensuring thread-safety and predictability.
    - **Performance**: Optimized for speed and memory efficiency by disabling garbage collection (`gc=False`)
      and using immutable data structures. This reduces overhead and can significantly boost performance,
      especially when handling many instances.
    - **Compactness**: Stores file content in memory as bytes, leading to fast access and manipulation.
      The absence of mutable state allows for leaner objects.
    - **Garbage Collection**: By setting `gc=False`, the class instances are excluded from garbage collection tracking.
      This improves performance when creating many small objects but requires careful management of resources.
    - **Compression Support**: Provides methods for compressing and decompressing file contents using gzip,
      returning new `RawFile` instances without altering the original data.
    - **Versatile Creation Methods**: Offers multiple class methods to create `RawFile` instances from various sources,
      such as file paths, bytes, base64 strings, strings, streams, URLs, and cloud storage services.

    **Important Notes:**

    - **Memory Usage**: Since the entire file content is stored in memory, handling very large files may lead
      to high memory consumption. Ensure that file sizes are manageable within the available system memory.
    - **Resource Management**: As garbage collection is disabled, it's crucial to manage resources appropriately.
      While the class is designed to be immutable and not require cleanup, be cautious when handling external resources.
    - **Thread-Safety**: Immutability ensures that instances of `RawFile` are inherently thread-safe.

    **Example Usage:**

    ```python
    # Create a RawFile instance from a file path
    raw_file = RawFile.from_file_path('example.pdf')

    # Access the file extension
    print(raw_file.extension)  # Output: FileExtension.PDF

    # Get the size of the file content
    print(raw_file.get_size())  # Output: Size of the file in bytes

    # Compute checksums
    md5_checksum = raw_file.compute_md5()
    sha256_checksum = raw_file.compute_sha256()

    # Save the content to a new file
    raw_file.save_to_file('copy_of_example.pdf')

    # Compress the file content
    compressed_file = raw_file.compress()

    # Decompress the file content
    decompressed_file = compressed_file.decompress()
    ```

    **Methods Overview:**

    - Creation:
      - `from_file_path(cls, file_path: str)`: Create from a file path.
      - `from_bytes(cls, data: bytes, extension: FileExtension)`: Create from bytes.
      - `from_base64(cls, b64_string: str, extension: FileExtension)`: Create from a base64 string.
      - `from_string(cls, content: str, extension: FileExtension, encoding: str = "utf-8")`: Create from a string.
      - `from_stream(cls, stream: BinaryIO, extension: FileExtension)`: Create from a binary stream.
      - `from_url(cls, url: str, ...)`: Create from a URL.
      - `from_s3(cls, bucket_name: str, object_key: str, extension: Optional[FileExtension] = None)`: Create from Amazon S3.
      - `from_azure_blob(cls, connection_string: str, container_name: str, blob_name: str, extension: Optional[FileExtension] = None)`: Create from Azure Blob Storage.
      - `from_gcs(cls, bucket_name: str, blob_name: str, extension: Optional[FileExtension] = None)`: Create from Google Cloud Storage.
      - `from_zip(cls, zip_file_path: str, inner_file_path: str, extension: Optional[FileExtension] = None)`: Create from a file within a ZIP archive.
      - `from_stdin(cls, extension: FileExtension)`: Create from standard input.

    - Utilities:
      - `save_to_file(self, file_path: str)`: Save content to a file.
      - `get_size(self) -> int`: Get the size of the content in bytes.
      - `compute_md5(self) -> str`: Compute MD5 checksum.
      - `compute_sha256(self) -> str`: Compute SHA256 checksum.
      - `get_mime_type(self) -> str`: Get MIME type based on the file extension.
      - `compress(self) -> RawFile`: Compress content using gzip.
      - `decompress(self) -> RawFile`: Decompress gzip-compressed content.
      - `read_async(self) -> bytes`: Asynchronously read the content.

    **Immutability Enforcement:**

    - The class is decorated with `msgspec.Struct` and `frozen=True`, which makes all instances immutable.
    - Any method that would traditionally modify the instance returns a new `RawFile` instance instead.
    - This design ensures that the original data remains unchanged, promoting safer and more predictable code.

    **Performance Considerations:**

    - **No Garbage Collection Overhead**: By setting `gc=False`, instances are not tracked by the garbage collector, reducing overhead.
      This is suitable when instances do not contain cyclic references.
    - **Optimized Data Structures**: Using immutable bytes and avoiding mutable state enhances performance and reduces memory footprint.
    - **Fast Access**: In-memory storage allows for rapid access and manipulation of file content.

    **Garbage Collection and Resource Management:**

    - While garbage collection is disabled for instances, Python's reference counting will still deallocate objects when they are no longer in use.
    - Be mindful when working with external resources (e.g., open files or network connections) to ensure they are properly closed.
    - Since `RawFile` instances hold data in memory, they are automatically cleaned up when references are removed.

    **Thread-Safety:**

    - Immutable objects are inherently thread-safe because their state cannot change after creation.
    - `RawFile` instances can be shared across threads without the need for synchronization mechanisms.

    **Compression Level:**

    - The `compress` and `decompress` methods use gzip with default compression levels.
    - If you need to specify a compression level, you can modify the methods to accept a parameter for the compression level.

    **Extensibility:**

    - The `FileExtension` enum and content type mappings can be extended to support additional file types as needed.
    - Custom methods can be added to handle specific use cases or integrations with other services.

    **Examples of Creating `RawFile` Instances from Different Sources:**

    ```python
    # From bytes
    raw_file = RawFile.from_bytes(b"Hello, World!", FileExtension.TXT)

    # From a base64 string
    raw_file = RawFile.from_base64("SGVsbG8sIFdvcmxkIQ==", FileExtension.TXT)

    # From a URL
    raw_file = RawFile.from_url("https://example.com/data.json")

    # From Amazon S3
    raw_file = RawFile.from_s3("my-bucket", "path/to/object.json")

    # From Azure Blob Storage
    raw_file = RawFile.from_azure_blob("connection_string", "container", "blob.json")

    # From Google Cloud Storage
    raw_file = RawFile.from_gcs("my-bucket", "path/to/blob.json")

    # From standard input
    raw_file = RawFile.from_stdin(FileExtension.TXT)
    ```

    **Disclaimer:**

    - Ensure that all necessary dependencies are installed for methods that interface with external services.
    - Handle exceptions appropriately in production code, especially when dealing with I/O operations and network requests.
    """

    name: Annotated[
        str,
        msgspec.Meta(
            title="Name", description="The name of the file", examples=["example.pdf"]
        ),
    ]
    contents: bytes
    extension: FileExtension

    @classmethod
    def from_file_path(cls: type[RawFile], file_path: str) -> RawFile:
        path = Path(file_path)
        if not path.exists():
            raise FileNotFoundError(f"File not found at {file_path}")

        if not path.is_file():
            raise ValueError(f"{file_path} is not a file")

        with open(file_path, "rb") as f:
            data = f.read()

        return cls(
            name=path.name,
            contents=data,
            extension=FileExtension(path.suffix.lstrip(".")),
        )

    @classmethod
    def from_bytes(cls, data: bytes, name: str, extension: FileExtension) -> RawFile:
        return cls(name=name, contents=data, extension=extension)

    @classmethod
    def from_base64(
        cls, b64_string: str, name: str, extension: FileExtension
    ) -> RawFile:
        data = base64.b64decode(b64_string)
        return cls.from_bytes(data=data, name=name, extension=extension)

    @classmethod
    def from_string(
        cls, content: str, name: str, extension: FileExtension, encoding: str = "utf-8"
    ) -> RawFile:
        data = content.encode(encoding)
        return cls.from_bytes(data=data, name=name, extension=extension)

    @classmethod
    def from_stream(
        cls, stream: BinaryIO, name: str, extension: FileExtension
    ) -> RawFile:
        data = stream.read()
        return cls(name=name, contents=data, extension=extension)

    @overload
    @classmethod
    def from_litestar_upload_file(
        cls: type[RawFile], file: LitestarUploadFile, is_zip: Literal[False] = False
    ) -> RawFile: ...

    @overload
    @classmethod
    def from_litestar_upload_file(
        cls: type[RawFile], file: LitestarUploadFile, is_zip: Literal[True]
    ) -> Sequence[RawFile]: ...

    @classmethod
    @ensure_module_installed("litestar", "litestar")
    def from_litestar_upload_file(
        cls: type[RawFile], file: LitestarUploadFile, is_zip: bool = False
    ) -> RawFile | Sequence[RawFile]:
        extension: Optional[FileExtension] = cls._get_extension_from_content_type(
            file.content_type
        )

        if extension is None:
            raise ValueError(f"{file.content_type} is not a supported file type yet.")

        if extension == FileExtension.ZIP:
            with tempfile.TemporaryDirectory() as temp_dir:
                file_contents = file.file.read()
                temp_file_path = Path(temp_dir) / file.filename
                with open(temp_file_path, "wb") as f:
                    f.write(file_contents)

                return cls.from_zip(str(temp_file_path))

        data = file.file.read()
        return cls(name=file.filename, contents=data, extension=extension)

    @classmethod
    @ensure_module_installed("fastapi", "fastapi")
    def from_fastapi_upload_file(
        cls: type[RawFile], file: FastAPIUploadFile
    ) -> RawFile:
        if file.content_type is None:
            raise ValueError("The content type of the file is missing.")

        extension: Optional[FileExtension] = cls._get_extension_from_content_type(
            file.content_type
        )

        if extension is None:
            raise ValueError(f"{file.content_type} is not a supported file type yet.")

        data = file.file.read()
        return cls(name=file.filename, contents=data, extension=extension)

    @classmethod
    def from_url(
        cls: type[Self],
        url: str,
        *,
        params: Optional[_Params] = None,
        data: Optional[_Data] = None,
        headers: Optional[_HeadersMapping] = None,
        cookies: Optional[CookieJar | _TextMapping] = None,
        files: Optional[_Files] = None,
        auth: Optional[_Auth] = None,
        timeout: Optional[_Timeout] = None,
        allow_redirects: bool = False,
        proxies: Optional[_TextMapping] = None,
        hooks: Optional[_HooksInput] = None,
        stream: Optional[bool] = None,
        verify: Optional[_Verify] = None,
        cert: Optional[_Cert] = None,
        json: Optional[Incomplete] = None,
        extension: Optional[FileExtension] = None,
    ) -> RawFile:
        response: requests.Response = requests.get(
            url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )
        response.encoding = "utf-8"

        response_content: bytes | Any = response.content

        if not isinstance(response_content, bytes):
            data = str(data).encode("utf-8")

        file_extension = extension or (
            cls._get_extension_from_content_type(
                response.headers.get("Content-Type", "").split(";")[0]
            )
            or FileExtension.HTML
        )

        return cls(name=url, contents=response_content, extension=file_extension)

    @ensure_module_installed("boto3", "boto3")
    @classmethod
    def from_s3(
        cls,
        bucket_name: str,
        object_key: str,
        extension: Optional[FileExtension] = None,
    ) -> RawFile:
        import boto3

        s3 = boto3.client("s3")

        if not extension:
            ext = Path(object_key).suffix.lstrip(".")
            if ext.upper() in FileExtension.__members__:
                extension = FileExtension[ext.upper()]
            else:
                head_object = s3.head_object(Bucket=bucket_name, Key=object_key)
                content_type = head_object.get("ContentType", "")
                extension = cls._get_extension_from_content_type(content_type)

        if not extension:
            raise ValueError(
                "Unable to determine the file extension. Please specify it explicitly."
            )

        obj = s3.get_object(Bucket=bucket_name, Key=object_key)
        data = obj["Body"].read()

        return cls(name=bucket_name, contents=data, extension=extension)

    @ensure_module_installed("azure.storage.blob", "azure-storage-blob")
    @classmethod
    def from_azure_blob(
        cls,
        connection_string: str,
        container_name: str,
        blob_name: str,
        extension: Optional[FileExtension] = None,
    ) -> RawFile:
        from azure.storage.blob import BlobServiceClient

        blob_service_client = BlobServiceClient.from_connection_string(
            connection_string
        )
        blob_client = blob_service_client.get_blob_client(
            container=container_name, blob=blob_name
        )

        if not extension:
            ext = Path(blob_name).suffix.lstrip(".")
            if ext.upper() in FileExtension.__members__:
                extension = FileExtension[ext.upper()]
            else:
                properties = blob_client.get_blob_properties()
                content_type = properties.content_settings.content_type
                if content_type is None:
                    raise ValueError(
                        "Unable to determine the file extension. Please specify it explicitly."
                    )

                extension = cls._get_extension_from_content_type(content_type)

        if not extension:
            raise ValueError(
                "Unable to determine the file extension. Please specify it explicitly."
            )

        stream = blob_client.download_blob()
        data = stream.readall()

        return cls(name=container_name, contents=data, extension=extension)

    @ensure_module_installed("google.cloud.storage", "google-cloud-storage")
    @classmethod
    def from_gcs(
        cls, bucket_name: str, blob_name: str, extension: Optional[FileExtension] = None
    ) -> RawFile:
        from google.cloud.storage import Client

        client = Client()
        bucket = client.bucket(bucket_name)
        blob = bucket.blob(blob_name)

        if not extension:
            ext = Path(blob_name).suffix.lstrip(".")
            if ext.upper() in FileExtension.__members__:
                extension = FileExtension[ext.upper()]
            else:
                blob.reload()
                content_type = blob.content_type
                extension = cls._get_extension_from_content_type(content_type)

        if not extension:
            raise ValueError(
                "Unable to determine the file extension. Please specify it explicitly."
            )

        data = blob.download_as_bytes()

        return cls(name=bucket_name, contents=data, extension=extension)

    @overload
    @classmethod
    def from_zip(
        cls,
        zip_file_path: str,
        inner_file_path: None = None,
    ) -> Sequence[RawFile]: ...

    @overload
    @classmethod
    def from_zip(
        cls,
        zip_file_path: str,
        inner_file_path: str,
    ) -> RawFile: ...

    @classmethod
    def from_zip(
        cls,
        zip_file_path: str,
        inner_file_path: Optional[str] = None,
    ) -> Sequence[RawFile] | RawFile:
        with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
            if inner_file_path:
                try:
                    with zip_ref.open(inner_file_path) as file:
                        data = file.read()
                        file_extension = FileExtension(
                            Path(inner_file_path).suffix.lstrip(".")
                        )
                        return cls(
                            name=inner_file_path,
                            contents=data,
                            extension=file_extension,
                        )
                except KeyError:
                    raise FileNotFoundError(
                        f"File '{inner_file_path}' not found in zip archive '{zip_file_path}'"
                    )

            raw_files = []
            for inner_file in zip_ref.namelist():
                with zip_ref.open(inner_file) as file:
                    data = file.read()
                    file_extension = FileExtension(Path(inner_file).suffix.lstrip("."))
                    raw_files.append(
                        cls(name=inner_file, contents=data, extension=file_extension)
                    )
            return raw_files

    @classmethod
    def from_database_blob(cls, blob_data: bytes, extension: FileExtension) -> RawFile:
        return cls.from_bytes(name="database_blob", data=blob_data, extension=extension)

    @classmethod
    def from_stdin(cls, extension: FileExtension) -> RawFile:
        data = sys.stdin.buffer.read()
        return cls.from_bytes(name="stdin", data=data, extension=extension)

    @classmethod
    def from_ftp(
        cls,
        host: str,
        filepath: str,
        username: str = "",
        password: str = "",
        extension: Optional[FileExtension] = None,
    ) -> RawFile:
        import ftplib

        ftp = ftplib.FTP(host)
        ftp.login(user=username, passwd=password)
        data = bytearray()
        ftp.retrbinary(f"RETR {filepath}", data.extend)
        ftp.quit()
        if not extension:
            extension = FileExtension(Path(filepath).suffix.lstrip("."))
        return cls(name=filepath, contents=bytes(data), extension=extension)

    def save_to_file(self, file_path: str) -> None:
        with open(file_path, "wb") as f:
            f.write(self.contents)

    def get_size(self) -> int:
        return len(self.contents)

    def compute_md5(self) -> str:
        md5 = hashlib.md5()
        md5.update(self.contents)
        return md5.hexdigest()

    def compute_sha256(self) -> str:
        sha256 = hashlib.sha256()
        sha256.update(self.contents)
        return sha256.hexdigest()

    def get_mime_type(self) -> str:
        mime_type, _ = mimetypes.guess_type(f"file.{self.extension}")
        return mime_type or "application/octet-stream"

    def compress(self) -> RawFile:
        import gzip

        compressed_data = gzip.compress(self.contents)
        return RawFile(
            name="compressed.zip", contents=compressed_data, extension=self.extension
        )  # TODO

    def decompress(self) -> RawFile:
        import gzip

        decompressed_data = gzip.decompress(self.contents)
        return RawFile(
            name=self.name, contents=decompressed_data, extension=self.extension
        )

    async def read_async(self) -> bytes:
        return self.contents

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass  # Nothing to close since we're using bytes

    def __del__(self):
        pass  # No cleanup needed

    @staticmethod
    def _get_extension_from_content_type(content_type: str) -> Optional[FileExtension]:
        content_type_map = {
            "application/pdf": FileExtension.PDF,
            "application/json": FileExtension.JSON,
            "image/png": FileExtension.PNG,
            "image/jpeg": FileExtension.JPEG,
            "image/jpg": FileExtension.JPG,
            "text/html": FileExtension.HTML,
            "text/plain": FileExtension.TXT,
            "application/x-zip-compressed": FileExtension.ZIP,
            # Add more mappings as needed
        }
        return content_type_map.get(content_type, None)



"""Builder module for providing semantic hints and easier to understand code."""

from __future__ import annotations
from typing import Annotated, Any
import msgspec


class DynamicDict(msgspec.Struct):
    """Constructor parameters for a class."""

    params: Annotated[
        dict[str, Any],
        msgspec.Meta(title="Params", description="Parameters of the constructor"),
    ] = msgspec.field()

    @classmethod
    def having(cls, key: str, /, equals_to: Any) -> DynamicDict:
        return DynamicDict(params={key: equals_to})

    def as_well_as(self, key: str, /, equals_to: Any) -> DynamicDict:
        self.params.update({key: equals_to})
        return self

    def also(self, key: str, /, equals_to: Any, last: bool = False) -> DynamicDict:
        self.params.update({key: equals_to})
        return self

    def at_last(self, key: str, /, equals_to: Any) -> dict[str, Any]:
        self.params.update({key: equals_to})
        return self.params



"""
This module contains marker decorators for classes.

These markers serve two purposes:

1. To annotate classes with their intended roles or characteristics, enhancing
    code readability and maintainability.
2. To be used by the `generate_code_metrics` function (in a separate module)
    to count the occurrences of each marker throughout the codebase.
"""

MARKER_COUNTS: dict[str, set[type]] = {}


def _register_marker(marker_name: str, cls: type) -> None:
    """Registers that a marker was applied to a class."""
    if marker_name not in MARKER_COUNTS:
        MARKER_COUNTS[marker_name] = set()
    MARKER_COUNTS[marker_name].add(cls)


def Factory[T](cls: type[T]) -> type[T]:
    """Marks a class as a Factory."""
    _register_marker("Factory", cls)
    return cls


def Singleton[T](cls: type[T]) -> type[T]:
    """Marks a class as a Singleton."""
    _register_marker("Singleton", cls)
    return cls


def Observer[T](cls: type[T]) -> type[T]:
    """Marks a class as an Observer."""
    _register_marker("Observer", cls)
    return cls


def Decorator[T](cls: type[T]) -> type[T]:
    """Marks a class as a Decorator."""
    _register_marker("Decorator", cls)
    return cls


def Adapter[T](cls: type[T]) -> type[T]:
    """Marks a class as an Adapter."""
    _register_marker("Adapter", cls)
    return cls



from typing import Any, Optional, TypeVar
import msgspec

S = TypeVar("S", bound=msgspec.Struct)


def structify(
    data: dict[str, Any],
    struct: type[S],
    decoder: Optional[msgspec.json.Decoder[S]] = None,
    encoder: Optional[msgspec.json.Encoder] = None,
) -> S:
    """Create a struct from a dictionary"""
    decoder = decoder or msgspec.json.Decoder(type=struct)
    encoder = encoder or msgspec.json.Encoder()
    _struct: S = decoder.decode(encoder.encode(data))
    return _struct


def dictify(
    struct: msgspec.Struct,
    encoder: Optional[msgspec.json.Encoder] = None,
    decoder: Optional[msgspec.json.Decoder[dict[str, Any]]] = None,
) -> dict[str, Any]:
    """Create a dictionary from a struct"""
    encoder = encoder or msgspec.json.Encoder()
    decoder = decoder or msgspec.json.Decoder(type=dict)
    return decoder.decode(encoder.encode(struct))



import ast
import os
from jinja2 import Environment, FileSystemLoader


def find_python_files(directory: str) -> list[str]:
    """Recursively finds all Python files in a directory."""
    python_files: list[str] = []
    for root, _, files in os.walk(directory):
        for file in files:
            if file.endswith(".py"):
                python_files.append(os.path.join(root, file))
    return python_files


def parse_file_for_markers(filepath: str) -> dict[str, set[str]]:
    """Parses a Python file to find classes and applied marker decorators."""
    with open(filepath, "r") as source:
        tree = ast.parse(source.read())

    file_marker_counts: dict[str, set[str]] = {}

    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            class_name = node.name
            applied_markers: set[str] = set()
            for decorator in node.decorator_list:
                if isinstance(decorator, ast.Name):
                    applied_markers.add(decorator.id)
                elif isinstance(decorator, ast.Call) and isinstance(
                    decorator.func, ast.Name
                ):
                    applied_markers.add(
                        decorator.func.id
                    )  # Handle decorators with arguments if needed

            if applied_markers:
                file_marker_counts[class_name] = applied_markers

    return file_marker_counts


def generate_metrics_data(directory: str) -> dict[str, int]:
    """Generates the metrics data by scanning the codebase."""
    all_files = find_python_files(directory)
    marker_usage: dict[str, int] = {}

    for filepath in all_files:
        file_marker_data = parse_file_for_markers(filepath)
        for class_name, markers in file_marker_data.items():
            for marker in markers:
                marker_usage[marker] = marker_usage.get(marker, 0) + 1

    return marker_usage


def generate_html_report(metrics: dict[str, int]) -> str:
    """Generates an HTML report with a bar chart using Tailwind CSS and Chart.js."""
    env = Environment(loader=FileSystemLoader("."))
    template = env.from_string("""
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <script src="https://cdn.tailwindcss.com"></script>
        <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/all.min.css" integrity="sha512-9usAa10IRO0HhonpyAIVpjrylPvoDwiPUiKdWk5t3PyolY1cOd4DSE0Ga+ri4AuTroPR5aQvXU9xC6qOPnzFeg==" crossorigin="anonymous" referrerpolicy="no-referrer" />
        <script src="https://cdn.jsdelivr.net/npm/chart.js"></script>
        <title>Code Metrics Report</title>
    </head>
    <body class="bg-gradient-to-br from-gray-100 to-blue-50 dark:bg-gradient-to-br dark:from-gray-800 dark:to-gray-900">
        <div class="container mx-auto p-8 shadow-xl rounded-lg bg-white dark:bg-gray-800">
            <div class="text-center mb-8">
                <h1 class="text-4xl font-extrabold text-gray-800 dark:text-white mb-4">
                    <i class="fas fa-chart-line mr-2"></i> Interactive Code Metrics Dashboard
                </h1>
                <p class="text-gray-600 dark:text-gray-400 italic">Visualizing marker adoption across your codebase.</p>
            </div>
            <div class="mb-8">
                <canvas id="markerChart"></canvas>
            </div>
            <div class="overflow-x-auto">
                <table class="min-w-full divide-y divide-gray-200 dark:divide-gray-700 shadow overflow-hidden rounded-md">
                    <thead class="bg-gray-50 dark:bg-gray-700">
                        <tr>
                            <th scope="col" class="px-6 py-3 text-left text-xs font-semibold text-gray-500 dark:text-gray-300 uppercase tracking-wider">
                                <i class="fas fa-tags mr-1"></i> Marker
                            </th>
                            <th scope="col" class="px-6 py-3 text-left text-xs font-semibold text-gray-500 dark:text-gray-300 uppercase tracking-wider">
                                <i class="fas fa-sort-numeric-up-alt mr-1"></i> Count
                            </th>
                        </tr>
                    </thead>
                    <tbody class="bg-white dark:bg-gray-800 divide-y divide-gray-200 dark:divide-gray-700">
                        {% for marker, count in metrics.items() %}
                        <tr>
                            <td class="px-6 py-4 whitespace-nowrap text-sm text-gray-900 dark:text-gray-100 font-medium">{{ marker }}</td>
                            <td class="px-6 py-4 whitespace-nowrap text-sm text-gray-700 dark:text-gray-400">{{ count }}</td>
                        </tr>
                        {% endfor %}
                    </tbody>
                </table>
            </div>
        </div>
        <script>
            const metrics = {{ metrics|tojson }};
            const ctx = document.getElementById('markerChart').getContext('2d');
            const markerChart = new Chart(ctx, {
                type: 'bar',
                data: {
                    labels: Object.keys(metrics),
                    datasets: [{
                        label: 'Marker Usage Count',
                        data: Object.values(metrics),
                        backgroundColor: 'rgba(54, 162, 235, 0.8)',
                        borderColor: 'rgba(54, 162, 235, 1)',
                        borderWidth: 1
                    }]
                },
                options: {
                    responsive: true,
                    maintainAspectRatio: false,
                    scales: {
                        y: {
                            beginAtZero: true,
                            title: {
                                display: true,
                                text: 'Count'
                            }
                        },
                        x: {
                            title: {
                                display: true,
                                text: 'Markers'
                            }
                        }
                    },
                    plugins: {
                        title: {
                            display: true,
                            text: 'Marker Usage Distribution',
                            font: {
                                size: 18
                            }
                        },
                        legend: {
                            display: false,
                        }
                    }
                }
            });
        </script>
    </body>
    </html>
    """)
    return template.render(metrics=metrics)


def save_html_report(html_content: str, filepath: str) -> None:
    """Saves the HTML content to a file."""
    with open(filepath, "w", encoding="utf-8") as f:
        f.write(html_content)


def generate_code_metrics(codebase_directory: str) -> str:
    """
    Generates code metrics and creates an HTML report, saving it to a file.
    """
    metrics_data = generate_metrics_data(codebase_directory)
    html_report = generate_html_report(metrics_data)
    report_filepath = "code_metrics_report.html"
    save_html_report(html_report, report_filepath)
    return report_filepath


def example() -> None:
    codebase_path = "my_codebase"  # Replace with the actual path to your codebase
    example_file_path = os.path.join(codebase_path, "example.py")

    if not os.path.exists(codebase_path):
        os.makedirs(codebase_path)
    with open(example_file_path, "w") as f:
        f.write("""
from markers import Factory, Singleton, Observer

@Factory
class MyFactory:
    pass

@Singleton
class MySingleton:
    pass

@Observer
class MyObserver:
    pass
        """)

    report_filepath = generate_code_metrics(codebase_path)
    full_report_path = os.path.abspath(report_filepath)
    print(f"Report saved to: {full_report_path}")
    print(f"Open the file '{report_filepath}' in your web browser to view the report.")

    # Delete the example.py file
    if os.path.exists(example_file_path):
        os.remove(example_file_path)
        print(f"Deleted temporary file: {example_file_path}")


if __name__ == "__main__":
    example()



from __future__ import annotations

import collections.abc  # Required for runtime type checks
from typing import (
    Any,
    Callable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    cast,
    overload,
)

import msgspec
from pydantic_core import core_schema

U = TypeVar("U")
K = TypeVar("K")
V = TypeVar("V")


class Maybe[T = object](msgspec.Struct, frozen=True):
    """
    A class that safely handles optional chaining for Python objects, emulating the `?.` operator
    found in languages like JavaScript. This allows for safe access to attributes and methods
    of objects that may be `None`, preventing `AttributeError` exceptions.

    **Usage Patterns:**

    1. **Type Annotation with Instance Creation:**
       ```python
       user_instance = User("Alice")
       maybe_user: Maybe[User] = Maybe(user_instance)
       ```

    2. **Handling Optional Values:**
       ```python
       maybe_none_user: Maybe[User] = Maybe(None)
       ```

    **Usage Examples:**

    ```python
    >>> # Type annotation with instance creation
    >>> user_instance = User("Alice")
    >>> maybe_user: Maybe[User] = Maybe(user_instance)
    >>> maybe_user.name.unwrap()
    'Alice'

    >>> # Handling None
    >>> maybe_none_user: Maybe[User] = Maybe(None)
    >>> maybe_none_user.name.unwrap()
    None

    >>> # Wrapping a callable
    >>> def greet(user: User) -> str:
    ...     return f"Hello, {user.name}!"
    >>> maybe_greet: Maybe[Callable[[User], str]] = Maybe(greet)
    >>> maybe_greet(user_instance).unwrap()
    'Hello, Alice!'

    >>> # Wrapping a non-callable
    >>> maybe_not_callable: Maybe[int] = Maybe(42)
    >>> maybe_not_callable("Test").unwrap()
    None

    >>> # Using map to transform the wrapped value
    >>> maybe_number: Maybe[int] = Maybe(10)
    >>> maybe_double: Maybe[int] = maybe_number.map(lambda x: x * 2)
    >>> maybe_double.unwrap()
    20

    >>> # Using with_default to provide fallback
    >>> maybe_none: Maybe[str] = Maybe(None)
    >>> maybe_none.with_default("Default Value")
    'Default Value'

    >>> # Using and_then for chaining
    >>> maybe_upper: Maybe[str] = maybe_user.and_then(lambda user: Maybe(user.name.upper()))
    >>> maybe_upper.unwrap()
    'ALICE'

    >>> # Iterating over a wrapped list
    >>> maybe_list: Maybe[list[int]] = Maybe([1, 2, 3])
    >>> list(maybe_list)
    [1, 2, 3]

    >>> # Accessing items with __getitem__
    >>> maybe_dict: Maybe[dict[str, int]] = Maybe({"a": 1, "b": 2})
    >>> maybe_dict["a"].unwrap()
    1

    >>> maybe_dict["c"].unwrap()
    None
    ```
    """

    obj: Optional[T] = None

    def __getattr__(self, attr: str) -> Maybe[object]:
        """
        Safely access an attribute of the wrapped object.

        Args:
            attr (str): The attribute name to access.

        Returns:
            Maybe[Any]: An instance of `Maybe` wrapping the attribute's value or `None`.

        Examples:
            >>> class User:
            ...     def __init__(self, name):
            ...         self.name = name
            >>> user = User("Alice")
            >>> maybe_user: Maybe[User] = Maybe(user)
            >>> maybe_user.name.unwrap()
            'Alice'

            >>> maybe_none: Maybe[User] = Maybe(None)
            >>> maybe_none.name.unwrap()
            None
        """
        if self.obj is None:
            return Maybe[object](None)
        try:
            return Maybe[object](getattr(self.obj, attr))
        except AttributeError:
            return Maybe[object](None)

    def __call__(self, *args: Any, **kwargs: Any) -> Maybe[object]:
        """
        Safely call the wrapped object if it's callable.

        Args:
            *args: Positional arguments for the callable.
            **kwargs: Keyword arguments for the callable.

        Returns:
            Maybe[Any]: An instance of `Maybe` wrapping the result of the call or `None`.

        Examples:
            >>> def greet(user: User) -> str:
            ...     return f"Hello, {user.name}!"
            >>> maybe_greet: Maybe[Callable[[User], str]] = Maybe(greet)
            >>> maybe_greet(user_instance).unwrap()
            'Hello, Alice!'

            >>> maybe_callable_none: Maybe[Callable[[User], str]] = Maybe(None)
            >>> maybe_callable_none(user_instance).unwrap()
            None

            >>> maybe_not_callable: Maybe[int] = Maybe(42)
            >>> maybe_not_callable("Test").unwrap()
            None
        """

        if self.obj is None or not callable(self.obj):
            return Maybe[object](obj=None)
        try:
            result = self.obj(*args, **kwargs)
            return Maybe[object](obj=result)
        except Exception:
            return Maybe[object](obj=None)

    def map(self, func: Callable[[T], U]) -> Maybe[U]:
        """
        Apply a function to the wrapped object if it's not `None`.

        Args:
            func (Callable[[T], U]): A callable that takes the wrapped object and returns a new value.

        Returns:
            Maybe[U]: An instance of `Maybe` wrapping the function's result or `None`.

        Examples:
            >>> maybe_number: Maybe[int] = Maybe(10)
            >>> maybe_double: Maybe[int] = maybe_number.map(lambda x: x * 2)
            >>> maybe_double.unwrap()
            20

            >>> maybe_none: Maybe[str] = Maybe(None)
            >>> maybe_none.map(lambda x: x.upper()).unwrap()
            None

            >>> def risky_division(x: int) -> float:
            ...     return 10 / x
            >>> maybe_zero: Maybe[int] = Maybe(0)
            >>> maybe_zero.map(risky_division).unwrap()
            None  # Due to ZeroDivisionError

            >>> maybe_five: Maybe[int] = Maybe(5)
            >>> maybe_five.map(risky_division).unwrap()
            2.0
        """
        if self.obj is None:
            return Maybe[U](obj=None)
        try:
            return Maybe[U](obj=func(self.obj))
        except Exception:
            return Maybe[U](obj=None)

    def unwrap(self) -> Optional[T]:
        """
        Retrieve the underlying object.

        Returns:
            Optional[T]: The wrapped object if not `None`; otherwise, `None`.

        Examples:
            >>> maybe = Maybe("Hello")
            >>> maybe.unwrap()
            'Hello'

            >>> maybe_none = Maybe(None)
            >>> maybe_none.unwrap()
            None
        """
        return self.obj

    def __bool__(self) -> bool:
        """
        Allow `Maybe` instances to be used in boolean contexts.

        Returns:
            bool: `True` if the wrapped object is truthy; `False` otherwise.

        Examples:
            >>> maybe_true: Maybe[int] = Maybe(5)
            >>> bool(maybe_true)
            True

            >>> maybe_false: Maybe[int] = Maybe(0)
            >>> bool(maybe_false)
            False

            >>> maybe_none: Maybe[int] = Maybe(None)
            >>> bool(maybe_none)
            False
        """
        return self.obj is not None

    def __eq__(self, other: object) -> bool:
        """
        Equality comparison between `Maybe` instances or with raw values.

        Args:
            other (object): Another `Maybe` instance or a raw value to compare with.

        Returns:
            bool: `True` if both wrapped objects are equal; `False` otherwise.

        Examples:
            >>> maybe1: Maybe[int] = Maybe(5)
            >>> maybe2: Maybe[int] = Maybe(5)
            >>> maybe3: Maybe[int] = Maybe(10)
            >>> maybe1 == maybe2
            True
            >>> maybe1 == maybe3
            False
            >>> maybe1 == 5
            True
            >>> maybe_none: Maybe[int] = Maybe(None)
            >>> maybe_none == None
            True
        """
        if isinstance(other, Maybe):
            return self._obj == other._obj
        return self._obj == other

    def __ne__(self, other: object) -> bool:
        """
        Non-equality comparison between `Maybe` instances or with raw values.

        Args:
            other (object): Another `Maybe` instance or a raw value to compare with.

        Returns:
            bool: `True` if both wrapped objects are not equal; `False` otherwise.

        Examples:
            >>> maybe1: Maybe[int] = Maybe(5)
            >>> maybe2: Maybe[int] = Maybe(5)
            >>> maybe3: Maybe[int] = Maybe(10)
            >>> maybe1 != maybe2
            False
            >>> maybe1 != maybe3
            True
            >>> maybe1 != 5
            False
            >>> maybe_none: Maybe[int] = Maybe(None)
            >>> maybe_none != None
            False
        """
        return not self.__eq__(other)

    def __iter__(self) -> Iterator[Any]:
        """
        Allow iteration over the wrapped object if it's iterable.

        Yields:
            Any: Items from the wrapped iterable or nothing if the wrapped object is `None`.

        Examples:
            >>> maybe_list: Maybe[list[int]] = Maybe([1, 2, 3])
            >>> list(maybe_list)
            [1, 2, 3]

            >>> maybe_string: Maybe[str] = Maybe("abc")
            >>> list(maybe_string)
            ['a', 'b', 'c']

            >>> maybe_none: Maybe[list[int]] = Maybe(None)
            >>> list(maybe_none)
            []
        """
        if self.obj is not None and isinstance(self.obj, collections.abc.Iterable):
            # We do a runtime ignore on the type of `iter(obj)` because T might not be known to be iterable.
            return iter(self.obj)
        return iter(())

    @overload
    def __getitem__(self: Maybe[Mapping[K, V]], key: K) -> Maybe[V]: ...

    @overload
    def __getitem__(self: Maybe[Sequence[V]], key: int) -> Maybe[V]: ...

    @overload
    def __getitem__(self, key: object) -> Maybe[object]: ...

    def __getitem__(self, key: object) -> Maybe[Any]:
        """
        Safely access an item by key/index if the wrapped object supports indexing.

        Args:
            key (Any): The key/index to access.

        Returns:
            Maybe[Any]: An instance of `Maybe` wrapping the item's value or `None`.

        Examples:
            >>> maybe_dict: Maybe[dict[str, int]] = Maybe({"a": 1, "b": 2})
            >>> maybe_dict["a"].unwrap()
            1

            >>> maybe_dict["c"].unwrap()
            None

            >>> maybe_list: Maybe[list[int]] = Maybe([10, 20, 30])
            >>> maybe_list[1].unwrap()
            20

            >>> maybe_none: Maybe[dict[str, int]] = Maybe(None)
            >>> maybe_none["a"].unwrap()
            None
        """
        if self.obj is None:
            return Maybe(None)

        # 1) If obj is a Mapping, key must be the type of mapping keys (K).
        if isinstance(self.obj, Mapping):
            try:
                return Maybe(self.obj[key])
            except (KeyError, TypeError):
                return Maybe(None)

        # 2) If obj is a Sequence, key must be int or slice.
        elif isinstance(self.obj, Sequence):
            if not isinstance(key, (int, slice)):
                return Maybe(None)
            try:
                return Maybe(self.obj[key])
            except (IndexError, TypeError):
                return Maybe(None)

        # 3) If it just has __getitem__, we can attempt calling it:
        elif hasattr(self.obj, "__getitem__"):
            try:
                return Maybe(cast(Mapping, self.obj)[key])
            except (IndexError, KeyError, TypeError, AttributeError):
                return Maybe(None)

        return Maybe(None)

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        """
        Define how pydantic should handle the Maybe class during serialization and deserialization.
        """
        # Extract the type argument T from Maybe[T]
        wrapped_type = source_type.__args__[0]

        # Get the schema for the wrapped type T
        wrapped_schema = handler(wrapped_type)

        # Define the validation function (accepts only 'value')
        def validate(value: Any) -> Maybe[T]:
            return cls(value)

        # Define the serialization function (accepts only 'value')
        def serialize(value: "Maybe[T]") -> Any:
            return value.unwrap()

        # Create and return the CoreSchema
        schema = core_schema.no_info_after_validator_function(
            validate,
            wrapped_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                serialize, info_arg=False
            ),
        )
        return schema

    def with_default(self, default: T) -> T:
        """
        Provide a default value if the wrapped object is `None`.

        Args:
            default (T): The default value to return if the wrapped object is `None`.

        Returns:
            T: The wrapped object if not `None`; otherwise, the default value.

        Examples:
            >>> maybe_none: Maybe[str] = Maybe(None)
            >>> maybe_none.with_default("Default Value")
            'Default Value'

            >>> maybe_value: Maybe[str] = Maybe("Actual Value")
            >>> maybe_value.with_default("Default Value")
            'Actual Value'
        """
        return self.obj if self.obj is not None else default

    def and_then(self, func: Callable[[T], Maybe[U]]) -> Maybe[U]:
        """
        Chain operations that return `Maybe` instances.

        Args:
            func (Callable[[T], Maybe[U]]): A callable that takes the wrapped object and returns a `Maybe` instance.

        Returns:
            Maybe[U]: The result of the callable or `Maybe(None)` if the wrapped object is `None`.

        Examples:
            >>> def to_upper(s: str) -> Maybe[str]:
            ...     return Maybe(s.upper())
            >>> maybe_str: Maybe[str] = Maybe("hello")
            >>> upper_optional: Maybe[str] = maybe_str.and_then(to_upper)
            >>> upper_optional.unwrap()
            'HELLO'

            >>> def reverse_string(s: str) -> Maybe[str]:
            ...     return Maybe(s[::-1])
            >>> chained_optional: Maybe[str] = maybe_str.and_then(to_upper).and_then(reverse_string)
            >>> chained_optional.unwrap()
            'OLLEH'

            >>> def to_none(s: str) -> Maybe[str]:
            ...     return Maybe(None)
            >>> chained_none: Maybe[str] = maybe_str.and_then(to_none)
            >>> chained_none.unwrap()
            None

            >>> maybe_initial_none: Maybe[str] = Maybe(None)
            >>> chained_none_initial: Maybe[str] = maybe_initial_none.and_then(to_upper)
            >>> chained_none_initial.unwrap()
            None
        """
        if self.obj is None:
            return Maybe[U](obj=None)
        try:
            return func(self.obj)
        except Exception:
            return Maybe[U](obj=None)



import asyncio
import functools
from typing import (
    Any,
    Awaitable,
    Callable,
    Optional,
    ParamSpec,
    TypeVar,
    Union,
    cast,  # Import cast
    overload,
)
import importlib.util
from functools import wraps
from rich.console import Console
from rich.panel import Panel
from aiocache import cached as aiocache_decorator  # type: ignore[import-untyped]

# Define type variables
P = ParamSpec("P")
R = TypeVar("R")


def is_coroutine_function(func: Callable[..., Any]) -> bool:
    return asyncio.iscoroutinefunction(func)


@overload
def pure(
    func: Callable[P, R],
    *,
    cached: bool = ...,
    maxsize: Optional[int] = ...,
    ttl: Optional[int] = ...,
) -> Callable[P, R]: ...


@overload
def pure(
    func: None = ...,
    *,
    cached: bool = ...,
    maxsize: Optional[int] = ...,
    ttl: Optional[int] = ...,
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...


def pure(
    func: Optional[Callable[P, Union[R, Awaitable[R]]]] = None,
    *,
    cached: bool = False,
    maxsize: Optional[int] = None,
    ttl: Optional[int] = None,
) -> Union[
    Callable[P, Union[R, Awaitable[R]]],
    Callable[
        [Callable[P, Union[R, Awaitable[R]]]], Callable[P, Union[R, Awaitable[R]]]
    ],
]:
    """
    Decorator to cache functions, supporting both synchronous and asynchronous functions.

    Args:
        func: The function to decorate. Can be None if used with arguments.
        cached: Whether to apply caching. Defaults to False.
        maxsize: Maximum size of the cache (for synchronous functions).
        ttl: Time-to-live for cache entries (for asynchronous functions).

    Raises:
        ValueError: If both maxsize and ttl are provided, or if cached is False but maxsize/ttl are set.
    """

    def decorator(
        inner_func: Callable[P, Union[R, Awaitable[R]]],
    ) -> Callable[P, Union[R, Awaitable[R]]]:
        # Validation
        if not cached:
            if maxsize is not None or ttl is not None:
                raise ValueError("Cannot set maxsize or ttl when cached is False.")
            return inner_func

        if maxsize is not None and ttl is not None:
            raise ValueError("Cannot set both maxsize and ttl at the same time.")

        if is_coroutine_function(inner_func):
            if maxsize is not None:
                raise ValueError("maxsize cannot be used with asynchronous functions.")
            # Apply aiocache decorator
            cache_kwargs: dict[str, Any] = {}
            if ttl is not None:
                cache_kwargs["ttl"] = ttl
            decorated_func = aiocache_decorator(**cache_kwargs)(inner_func)
            return cast(Callable[P, Union[R, Awaitable[R]]], decorated_func)
        else:
            if ttl is not None:
                raise ValueError("ttl cannot be used with synchronous functions.")
            # Apply functools.lru_cache decorator
            cache_kwargs = {}
            if maxsize is not None:
                cache_kwargs["maxsize"] = maxsize
            else:
                cache_kwargs["maxsize"] = 128  # Default maxsize
            decorated_func = functools.lru_cache(**cache_kwargs)(inner_func)
            return cast(Callable[P, Union[R, Awaitable[R]]], decorated_func)

    if func is None:
        return decorator
    else:
        return decorator(func)


def ensure_module_installed(
    module_name: str, package_name: Optional[str] = None
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    A decorator that ensures a Python module is installed before executing the function.

    Args:
        module_name: The import name of the module (e.g., 'vertexai.generative_models')
        package_name: Optional pip package name to install (e.g., 'google-cloud-vertexai').
    """

    def ensure():
        spec = importlib.util.find_spec(module_name)
        if spec is None:
            error_message = f"""
[red bold]📦 Import Error: Module Not Found[/red bold]

[yellow]Missing Module:[/yellow] {module_name}
"""
            if package_name:
                error_message += f"""
[green]To fix this, run one of these (according to your package manager):[/green]
[blue]pip install {package_name}[/blue]
[blue]uv add {package_name}[/blue]
[blue]poetry add {package_name}[/blue]
"""
            else:
                error_message += (
                    "\n[dim]Please ensure the module is installed correctly.[/dim]"
                )

            error_message += "\n[dim]For more help, visit our documentation at https://docs.example.com[/dim]"

            console = Console()
            console.print(
                Panel(
                    error_message,
                    title="❌ Error",
                    border_style="red",
                    padding=(1, 2),
                    expand=False,
                )
            )
            raise ImportError(f"Could not find module {module_name}")

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            # Check only when the function is called
            ensure()
            return func(*args, **kwargs)

        return wrapper

    return decorator


"""
The architecture your project needs.
"""

from msgspec import Meta, field

from .data import (
    AsyncRepository,
    CreateResult,
    DeleteResult,
    ReadAllResult,
    ReadResult,
    Repository,
    UpdateResult,
)
from .contracts import AsyncExecutable, Executable
from .utils.builders import DynamicDict
from .utils.markers import NotGiven

__all__: list[str] = [
    "field",
    "Meta",
    "Meta",
    "field",
    "AsyncRepository",
    "Repository",
    "ReadAllResult",
    "ReadResult",
    "CreateResult",
    "UpdateResult",
    "DeleteResult",
    "AsyncExecutable",
    "DynamicDict",
    "Executable",
    "NotGiven",
]



"""Creators"""

import importlib
import inspect
from typing import Any, Protocol, TypeGuard, runtime_checkable


@runtime_checkable
class _HasStructFields(Protocol):
    """Protocol for classes (like msgspec.Struct) that define __struct_fields__."""

    __struct_fields__: tuple[str, ...]


def _is_msgspec_struct(cls: type[object]) -> TypeGuard[type[_HasStructFields]]:
    """Runtime check to let static type checkers know cls has __struct_fields__."""
    return hasattr(cls, "__struct_fields__")


class DynamicInstanceCreator[T]:
    """
    A utility class responsible for creating an instance of a class with only the
    parameters that match its constructor or __struct_fields__ (for msgspec.Struct).
    """

    _cls: type[T]

    def __init__(self, cls: type[T]) -> None:
        self._cls = cls

    def create_instance(self, **kwargs: Any) -> T:
        """
        Creates an instance of `cls` by:
            1) Checking if it's a msgspec.Struct (via _is_msgspec_struct),
               and using __struct_fields__ if so.
            2) Otherwise, inspecting the __init__ signature to gather valid parameters.
            3) Filtering out any invalid parameters (unless the class accepts **kwargs).
            4) Calling the constructor exactly once with the filtered arguments.
        """

        if _is_msgspec_struct(self._cls):
            # We know from the type guard that __struct_fields__ is a tuple of str
            valid_parameters: set[str] = set(self._cls.__struct_fields__)
            # msgspec.Struct does not accept **kwargs by default, so filter strictly
            filtered_kwargs: dict[str, Any] = {
                k: v for k, v in kwargs.items() if k in valid_parameters
            }
        else:
            # Use inspect.signature for standard classes / dataclasses
            sig = inspect.signature(self._cls.__init__)
            valid_parameters = set()
            accepts_kwargs = False

            for param in sig.parameters.values():
                # Skip `self` since it's not a constructor argument
                if param.name == "self":
                    continue

                if param.kind == param.VAR_KEYWORD:
                    # If there's a **kwargs in the signature
                    accepts_kwargs = True
                else:
                    valid_parameters.add(param.name)

            if accepts_kwargs:
                # Class can handle arbitrary **kwargs, so no filtering needed
                filtered_kwargs = dict(kwargs)
            else:
                # Filter out anything not in valid_parameters
                filtered_kwargs = {
                    k: v for k, v in kwargs.items() if k in valid_parameters
                }

        # Instantiate the class exactly once with the filtered arguments
        try:
            instance = self._cls(**filtered_kwargs)
        except TypeError as e:
            raise TypeError(
                f"Could not instantiate {self._cls.__name__} with arguments: {filtered_kwargs}\n"
                f"Error: {e}"
            ) from e

        return instance


class ModuleClassLoader:
    """Loader for classes in different modules."""

    class_name: str

    def __init__(self, class_name: str) -> None:
        self.class_name = class_name

    def get_class_from_module(self, module_name: str) -> type[Any]:
        module = importlib.import_module(module_name)
        cls = getattr(module, self.class_name)
        return cls

    def create_instance_from_module(
        self,
        module_name: str,
        **kwargs: Any,
    ) -> Any:
        try:
            cls: type[Any] = self.get_class_from_module(module_name=module_name)

            # Filter kwargs to match the class constructor parameters
            sig = inspect.signature(cls)
            init_params = sig.parameters
            filtered_kwargs = {k: v for k, v in kwargs.items() if k in init_params}

            instance = cls(**filtered_kwargs)
            return instance
        except ImportError as e:
            raise ImportError(f"Could not import module '{module_name}': {e}") from e
        except AttributeError as e:
            raise AttributeError(
                f"Class '{self.class_name}' not found in module '{module_name}': {e}"
            ) from e
        except TypeError as e:
            raise TypeError(
                f"Could not instantiate {self.class_name} from module {module_name}: {e}"
            ) from e
        except Exception as e:
            raise Exception(
                f"An error occurred while creating instance from module: {e}"
            ) from e



from __future__ import annotations

import typing

import msgspec


class NotGiven(msgspec.Struct, frozen=True):  # type: ignore[call-arg]
    """
    `NotGiven` is a sentinel class used to distinguish between `None`, actual values,
    and values that have not been provided (i.e., absent).

    This class is particularly useful in scenarios where you need to differentiate
    between a parameter explicitly set to `None` and a parameter that was omitted
    or not provided at all.

    Attributes:
        None

    Example:
        ```python
        from typing import Optional

        def update_user(name: Optional[str] = NotGiven()):
            if NotGiven.is_absent(name):
                print("Name not provided.")
            elif name is None:
                print("Name set to None.")
            else:
                print(f"Name updated to {name}.")

        update_user()                  # Output: Name not provided.
        update_user(None)              # Output: Name set to None.
        update_user("Alice")           # Output: Name updated to Alice.
        ```
    """

    @classmethod
    def is_absent(cls, obj: typing.Any) -> bool:
        """
        Determine whether the provided object is an instance of `NotGiven`,
        indicating that the value was not provided.

        Args:
            obj (Any): The object to check.

        Returns:
            bool: `True` if `obj` is an instance of `NotGiven`, otherwise `False`.

        Example:
            ```python
            print(NotGiven.is_absent(NotGiven()))  # Output: True
            print(NotGiven.is_absent(None))         # Output: False
            print(NotGiven.is_absent("value"))      # Output: False
            ```
        """
        return isinstance(obj, NotGiven)



import asyncio
import threading
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Awaitable, Callable, Coroutine, ParamSpec, TypeVar

from .decorators import pure

P = ParamSpec("P")
R = TypeVar("R")


def file_get_contents(filename: str, cached: bool = False) -> str:
    """Read the contents of a filename and cache the result"""
    return (
        Path(filename).read_text()
        if not cached
        else _file_get_contents_cached(filename)
    )


@pure(cached=True)
def _file_get_contents_cached(filename: str) -> str:
    return Path(filename).read_text()


def run_sync[_T](func: Callable[..., Awaitable[_T]], *args, **kwargs) -> _T:
    """
    Runs a callable synchronously. If called from an async context in the main thread,
    it runs the callable in a new event loop in a separate thread. Otherwise, it
    runs the callable directly or using `run_coroutine_threadsafe`.

    Args:
        func: The callable to execute.
        *args: Positional arguments to pass to the callable.
        **kwargs: Keyword arguments to pass to the callable.

    Returns:
        The result of the callable.
    """

    async def _async_wrapper() -> _T:
        return await func(*args, **kwargs)

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(_async_wrapper())

    if threading.current_thread() is threading.main_thread():
        if not loop.is_running():
            return loop.run_until_complete(_async_wrapper())
        else:

            def run_in_new_loop():
                new_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(new_loop)
                try:
                    return new_loop.run_until_complete(_async_wrapper())
                finally:
                    new_loop.close()

            with ThreadPoolExecutor() as pool:
                future = pool.submit(run_in_new_loop)
                return future.result(30)
    else:
        return asyncio.run_coroutine_threadsafe(_async_wrapper(), loop).result()


def fire_and_forget(
    async_func: Callable[P, Coroutine[Any, Any, R]], *args: P.args, **kwargs: P.kwargs
) -> None:
    """
    Schedules the async_func to run in the existing event loop if one is running.
    Otherwise, it creates a new event loop and runs the coroutine to completion.

    This function does not wait for the coroutine to finish if a loop is already
    running ("fire-and-forget"). If no loop is detected in the current thread,
    it will block just long enough to run `async_func()` in a newly-created loop
    (which is closed immediately afterward).

    Args:
        async_func: The asynchronous function (coroutine) to run.
        *args: Positional arguments to pass to the coroutine.
        **kwargs: Keyword arguments to pass to the coroutine.
    """
    try:
        # Attempt to get a running loop in the current thread.
        loop = asyncio.get_running_loop()

        if loop.is_running():
            # We have a loop, and it's actively running. Schedule the coroutine
            # to run asynchronously (true fire-and-forget).
            loop.create_task(async_func(*args, **kwargs))
        else:
            # We have a loop object in this thread, but it's not actually running.
            # Run the coroutine to completion (blocking briefly).
            loop.run_until_complete(async_func(*args, **kwargs))
    except RuntimeError:
        # No event loop in the current thread -> create one and run the coroutine
        # immediately to completion, then close the loop.
        asyncio.run(async_func(*args, **kwargs))



"""Base interfaces for all repositories in the application.

Defines abstract base classes for synchronous and asynchronous repositories, facilitating
data access and manipulation across various storage mechanisms. These base classes ensure
a consistent interface for CRUD operations, supporting a clean architecture and
separation of concerns within the application.
"""

import abc
from typing import (
    Any,
    Generic,
    Optional,
    Protocol,
    Sequence,
    runtime_checkable,
    TypeVar,
)

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")


class CreateResult:
    """Result of a create operation, including the ID of the created entity."""

    uid: Optional[str]

    def __init__(self, uid: Optional[str] = None) -> None:
        """
        Initialize a CreateResult instance.

        Args:
            uid (Optional[str]): The unique identifier of the created entity. Defaults to None.
        """
        self.uid = uid


class ReadResult(Generic[T_co]):
    """Result of a read operation, including the retrieved entity."""

    entity: T_co

    def __init__(self, entity: T_co) -> None:
        """
        Initialize a ReadResult instance.

        Args:
            entity (Optional[T]): The entity retrieved by the operation. Defaults to None.
        """
        self.entity = entity


class ReadAllResult(Generic[T_co]):
    """Result of a read all operation, including a list of all entities."""

    entities: Sequence[T_co]

    def __init__(self, entities: list[T_co]) -> None:
        """
        Initialize a ReadAllResult instance.

        Args:
            entities (Optional[List[T]]): A list of all entities retrieved. Defaults to an empty list.
        """
        self.entities = entities


class UpdateResult:
    """Result of an update operation, including the number of affected records."""

    def __init__(self, affected_records: int) -> None:
        """
        Initialize an UpdateResult instance.

        Args:
            affected_records (int): The number of affected records. Defaults to 0.
        """
        self.affected_records = affected_records


class DeleteResult:
    """Result of a delete operation, including the number of affected records."""

    def __init__(self, affected_records: int) -> None:
        """
        Initialize a DeleteResult instance.

        Args:
            affected_records (int): The number of affected records. Defaults to 0.
        """
        self.affected_records = affected_records


@runtime_checkable
class AsyncRepository(Protocol[T]):
    @abc.abstractmethod
    async def create(
        self, entity: T, *, filters: Optional[dict[str, Any]] = None
    ) -> CreateResult: ...

    @abc.abstractmethod
    async def read(
        self, q: str, *, filters: Optional[dict[str, Any]] = None
    ) -> ReadResult[T]: ...

    @abc.abstractmethod
    async def read_all(
        self, *, filters: Optional[dict[str, Any]] = None
    ) -> ReadAllResult[T]: ...

    @abc.abstractmethod
    async def update(
        self, q: str, entity: T, *, filters: Optional[dict[str, Any]] = None
    ) -> UpdateResult: ...

    @abc.abstractmethod
    async def delete(
        self, q: str, *, filters: Optional[dict[str, Any]] = None
    ) -> DeleteResult: ...


@runtime_checkable
class Repository(Protocol[T]):
    @abc.abstractmethod
    def create(
        self, entity: T, *, filters: Optional[dict[str, Any]] = None
    ) -> CreateResult: ...

    @abc.abstractmethod
    def read(
        self, q: str, *, filters: Optional[dict[str, Any]] = None
    ) -> ReadResult[T]: ...

    @abc.abstractmethod
    def read_all(
        self, *, filters: Optional[dict[str, Any]] = None
    ) -> ReadAllResult[T]: ...

    @abc.abstractmethod
    def update(
        self, q: str, entity: T, *, filters: Optional[dict[str, Any]] = None
    ) -> UpdateResult: ...

    @abc.abstractmethod
    def delete(
        self, q: str, *, filters: Optional[dict[str, Any]] = None
    ) -> DeleteResult: ...



import unittest
from typing import Callable, Dict, List

from architecture.extensions import Maybe


class TestMaybe(unittest.TestCase):
    def test_initialization(self):
        maybe: Maybe[int] = Maybe(5)
        self.assertEqual(maybe.unwrap(), 5)
        maybe_none: Maybe[int] = Maybe(None)
        self.assertIsNone(maybe_none.unwrap())

    def test_getattr_existing_attribute(self):
        class TestClass:
            def __init__(self):
                self.value = 10

        obj = TestClass()
        maybe_obj: Maybe[TestClass] = Maybe(obj)
        self.assertEqual(maybe_obj.value.unwrap(), 10)

    def test_getattr_nonexistent_attribute(self):
        class TestClass:
            pass

        obj = TestClass()
        maybe_obj: Maybe[TestClass] = Maybe(obj)
        self.assertIsNone(maybe_obj.nonexistent.unwrap())

    def test_call_callable(self):
        def add_one(x: int) -> int:
            return x + 1

        maybe_callable: Maybe[Callable[[int], int]] = Maybe(add_one)
        self.assertEqual(maybe_callable(5).unwrap(), 6)

    def test_call_non_callable(self):
        maybe_non_callable: Maybe[int] = Maybe(10)
        self.assertIsNone(maybe_non_callable(5).unwrap())

    def test_call_none(self):
        maybe_none: Maybe[Callable[[int], int]] = Maybe(None)
        self.assertIsNone(maybe_none(5).unwrap())

    def test_map_existing_value(self):
        maybe: Maybe[int] = Maybe(5)
        self.assertEqual(maybe.map(lambda x: x * 2).unwrap(), 10)

    def test_map_none_value(self):
        maybe_none: Maybe[int] = Maybe(None)
        self.assertIsNone(maybe_none.map(lambda x: x * 2).unwrap())

    def test_map_with_exception(self):
        def risky_function(x: int) -> int:
            return 10 // x

        maybe_zero: Maybe[int] = Maybe(0)
        self.assertIsNone(maybe_zero.map(risky_function).unwrap())

        maybe_five: Maybe[int] = Maybe(5)
        self.assertEqual(maybe_five.map(risky_function).unwrap(), 2)

    def test_bool_true(self):
        maybe_true: Maybe[int] = Maybe(5)
        self.assertTrue(maybe_true)

    def test_bool_false_zero(self):
        maybe_false: Maybe[int] = Maybe(0)
        self.assertFalse(maybe_false)

    def test_bool_false_none(self):
        maybe_none: Maybe[int] = Maybe(None)
        self.assertFalse(maybe_none)

    def test_eq_same_maybe(self):
        maybe1: Maybe[int] = Maybe(5)
        maybe2: Maybe[int] = Maybe(5)
        maybe3: Maybe[int] = Maybe(10)
        self.assertEqual(maybe1, maybe2)
        self.assertNotEqual(maybe1, maybe3)

    def test_eq_with_value(self):
        maybe: Maybe[int] = Maybe(5)
        self.assertEqual(maybe, 5)
        self.assertNotEqual(maybe, 10)

    def test_eq_with_none(self):
        maybe_none: Maybe[int] = Maybe(None)
        self.assertEqual(maybe_none, None)
        self.assertNotEqual(maybe_none, 5)

    def test_hashable(self):
        maybe1: Maybe[int] = Maybe(5)
        maybe2: Maybe[int] = Maybe(5)
        self.assertEqual(hash(maybe1), hash(maybe2))

        maybe_none1: Maybe[None] = Maybe(None)
        maybe_none2: Maybe[None] = Maybe(None)
        self.assertEqual(hash(maybe_none1), hash(maybe_none2))

        maybe3: Maybe[int] = Maybe(10)
        self.assertNotEqual(hash(maybe1), hash(maybe3))

    def test_iterable_wrapped(self):
        maybe_list: Maybe[List[int]] = Maybe([1, 2, 3])
        self.assertEqual(list(maybe_list), [1, 2, 3])

        maybe_string: Maybe[str] = Maybe("abc")
        self.assertEqual(list(maybe_string), ["a", "b", "c"])

        maybe_none: Maybe[List[int]] = Maybe(None)
        self.assertEqual(list(maybe_none), [])

        maybe_non_iterable: Maybe[int] = Maybe(42)
        self.assertEqual(list(maybe_non_iterable), [])

    def test_getitem_existing_key(self):
        maybe_dict: Maybe[Dict[str, int]] = Maybe({"a": 1, "b": 2})
        self.assertEqual(maybe_dict["a"].unwrap(), 1)

    def test_getitem_nonexistent_key(self):
        maybe_dict: Maybe[Dict[str, int]] = Maybe({"a": 1, "b": 2})
        self.assertIsNone(maybe_dict["c"].unwrap())

    def test_getitem_existing_index(self):
        maybe_list: Maybe[List[int]] = Maybe([10, 20, 30])
        self.assertEqual(maybe_list[1].unwrap(), 20)

    def test_getitem_nonexistent_index(self):
        maybe_list: Maybe[List[int]] = Maybe([10, 20, 30])
        self.assertIsNone(maybe_list[5].unwrap())

    def test_getitem_on_none(self):
        maybe_none: Maybe[Dict[str, int]] = Maybe(None)
        self.assertIsNone(maybe_none["a"].unwrap())

    def test_with_default(self):
        maybe_none: Maybe[str] = Maybe(None)
        self.assertEqual(maybe_none.with_default("Default"), "Default")

        maybe_value: Maybe[str] = Maybe("Actual")
        self.assertEqual(maybe_value.with_default("Default"), "Actual")

    def test_and_then(self):
        def to_upper(s: str) -> Maybe[str]:
            return Maybe(s.upper())

        def reverse_string(s: str) -> Maybe[str]:
            return Maybe(s[::-1])

        maybe_str: Maybe[str] = Maybe("hello")
        upper_maybe: Maybe[str] = maybe_str.and_then(to_upper)
        self.assertEqual(upper_maybe.unwrap(), "HELLO")

        chained_optional: Maybe[str] = maybe_str.and_then(to_upper).and_then(
            reverse_string
        )
        self.assertEqual(chained_optional.unwrap(), "OLLEH")

        # Function that returns None
        def to_none(s: str) -> Maybe[str]:
            return Maybe(None)

        chained_none: Maybe[str] = maybe_str.and_then(to_none)
        self.assertIsNone(chained_none.unwrap())

        # Chaining on None
        maybe_initial_none: Maybe[str] = Maybe(None)
        chained_none_initial: Maybe[str] = maybe_initial_none.and_then(to_upper)
        self.assertIsNone(chained_none_initial.unwrap())


if __name__ == "__main__":
    unittest.main()



"""Módulo de configuração de logging do app"""

from __future__ import annotations

import logging
from rich.console import Console
from rich.theme import Theme
from rich.traceback import install as install_rich_traceback

# Install rich traceback handling
install_rich_traceback(show_locals=False)


class RichHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            level_name = record.levelname.lower()
            level_color = level_name

            log_message = (
                f"[{level_color}]{record.levelname:<8}[/{level_color}] | "
                f"[cyan]{self.formatTime(record)}[/cyan] | "
                f"[{level_color}]{record.getMessage()}[/{level_color}]"
            )

            console = Console(
                theme=Theme(
                    {
                        "debug": "dim cyan",
                        "info": "bold cyan",
                        "success": "bold green",
                        "warning": "bold yellow",
                        "error": "bold red",
                        "critical": "bold white on red",
                    }
                )
            )

            console.print(log_message, markup=True)

            # Print the exception traceback if any
            if record.exc_info:
                console.print_exception(
                    show_locals=False, width=100, extra_lines=3, word_wrap=True
                )
        except Exception:
            self.handleError(record)

    def formatTime(self, record, datefmt=None):
        return logging.Formatter("%(asctime)s").formatTime(record, datefmt)


def create_logger(name: str, level: int = logging.DEBUG) -> logging.Logger:
    """
    Creates a beautiful and rich logger for the application.

    Args:
        name (str): Nome do logger.
        level (int, optional): Nível de log para o logger. Defaults to logging.DEBUG.

    Returns:
        logging.Logger: Instância do logger configurado.
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)  # Set logger level using parameter

    if not logger.handlers:
        rich_handler = RichHandler()
        rich_handler.setLevel(logging.DEBUG)  # Set handler level
        logger.addHandler(rich_handler)

    return logger


# Use LoggerFactory to create a logger instance
logger = create_logger("global")  # Level is DEBUG by default



from __future__ import annotations

from typing import ParamSpec, Protocol, TypeVar, runtime_checkable

# Covariant TypeVars for return types
T_co_sync = TypeVar("T_co_sync", covariant=True)
T_co_async = TypeVar("T_co_async", covariant=True)

# Single ParamSpec for constructor parameters
P = ParamSpec("P")


@runtime_checkable
class Executable(Protocol[T_co_sync]):
    """Protocol for synchronous services within the application."""

    def execute(self) -> T_co_sync:
        """Performs the service's main operations."""
        ...


@runtime_checkable
class AsyncExecutable(Protocol[T_co_async]):
    """Protocol for asynchronous services within the application."""

    async def execute(self) -> T_co_async:
        """Performs the service's main operations asynchronously."""
        ...



from typing import Callable, Type, TypeVar

T = TypeVar("T")


def implements(_proto, /) -> Callable[[Type[T]], Type[T]]:
    """
    A decorator to indicate that a class implements a given Protocol.
    This decorator has no runtime impact and serves purely as documentation.
    """

    def decorator(cls: Type[T]) -> Type[T]:
        return cls

    return decorator



from .repositories import (
    AsyncRepository,
    CreateResult,
    DeleteResult,
    ReadAllResult,
    ReadResult,
    Repository,
    UpdateResult,
)

__all__: list[str] = [
    "AsyncRepository",
    "Repository",
    "CreateResult",
    "UpdateResult",
    "DeleteResult",
    "ReadAllResult",
    "ReadResult",
]
